@@ -27,8 +27,8 @@ from typing import Callable, Dict, List, Optional
 import typing
 
 import gi
-gi.require_versions({"Gfm": "0.1", "Grl": "0.3", "Tracker": "3.0"})
-from gi.repository import Gfm, Gio, Grl, GLib, GObject, Tracker
+gi.require_versions({"Grl": "0.3", "Tracker": "3.0"})
+from gi.repository import Grl, Gio, Gtk, GLib, GObject, Tracker
 
 from gnomemusic.asyncqueue import AsyncQueue
 from gnomemusic.corealbum import CoreAlbum
@@ -127,19 +127,19 @@ class GrlTrackerWrapper(GObject.GObject):
         self._notificationmanager: NotificationManager = (
             application.props.notificationmanager)
 
-        self._songs_search: Gfm.FilterListModel = Gfm.FilterListModel.new(
+        self._songs_search: Gtk.FilterListModel = Gtk.FilterListModel.new(
             self._songs_model)
-        self._songs_search.set_filter_func(lambda a: False)
+        self._songs_search.set_filter(Gtk.AnyFilter())
         cm.props.songs_search_proxy.append(self._songs_search)
 
-        self._albums_search: Gfm.FilterListModel = Gfm.FilterListModel.new(
+        self._albums_search: Gtk.FilterListModel = Gtk.FilterListModel.new(
             self._albums_model)
-        self._albums_search.set_filter_func(lambda a: False)
+        self._albums_search.set_filter(Gtk.AnyFilter())
         cm.props.albums_search_proxy.append(self._albums_search)
 
-        self._artists_search: Gfm.FilterListModel = Gfm.FilterListModel.new(
+        self._artists_search: Gtk.FilterListModel = Gtk.FilterListModel.new(
             self._artists_model)
-        self._artists_search.set_filter_func(lambda a: False)
+        self._artists_search.set_filter(Gtk.AnyFilter())
         cm.props.artists_search_proxy.append(self._artists_search)
 
         self._fast_options: Grl.OperationOptions = Grl.OperationOptions()
@@ -409,10 +409,20 @@ class GrlTrackerWrapper(GObject.GObject):
                     self._songs_model.remove(idx)
                     break
 
-    def _song_media_query(self, ids: List[str]) -> str:
-        media_ids = ", ".join([f"<{media_id}>" for media_id in ids])
+    def _song_media_query(self, ids: Optional[List[str]] = None) -> str:
+        """Returns a songs query string
 
-        query = """
+        :param list ids: List of Media ids to filter by or None
+        """
+        songs_filter = ""
+        if ids is not None:
+            media_ids = ", ".join([f"<{media_id}>" for media_id in ids])
+            songs_filter = f"FILTER ( ?song in ( {media_ids} ) )"
+
+        location_filter = self._tracker_wrapper.location_filter()
+        media_type = int(Grl.MediaType.AUDIO)
+        miner_fs_busname = self._tracker_wrapper.props.miner_fs_busname
+        query = " ".join(f"""
         SELECT
             ?type ?urn ?title ?id ?url
             ?artist ?album
@@ -420,11 +430,11 @@ class GrlTrackerWrapper(GObject.GObject):
             ?albumDiscNumber
             nie:usageCounter(?urn) AS ?playCount
             ?tag AS ?favorite
-        WHERE {
-            SERVICE <dbus:%(miner_fs_busname)s> {
-                GRAPH tracker:Audio {
+        WHERE {{
+            SERVICE <dbus:{miner_fs_busname}> {{
+                GRAPH tracker:Audio {{
                     SELECT DISTINCT
-                        %(media_type)s AS ?type
+                        {media_type} AS ?type
                         ?song AS ?urn
                         nie:title(?song) AS ?title
                         ?song AS ?id
@@ -435,24 +445,19 @@ class GrlTrackerWrapper(GObject.GObject):
                         nmm:trackNumber(?song) AS ?trackNumber
                         nmm:setNumber(nmm:musicAlbumDisc(?song))
                             AS ?albumDiscNumber
-                    WHERE {
+                    WHERE {{
                         ?song a nmm:MusicPiece .
-                        FILTER ( ?song in ( %(media_ids)s ) )
-                        %(location_filter)s
-                    }
-                }
-            }
-            OPTIONAL {
+                        {songs_filter}
+                        {location_filter}
+                    }}
+                }}
+            }}
+            OPTIONAL {{
                 ?urn nao:hasTag ?tag .
                 FILTER (?tag = nao:predefined-tag-favorite)
-            }
-        }
-        """.replace('\n', ' ').strip() % {
-            "miner_fs_busname": self._tracker_wrapper.props.miner_fs_busname,
-            "media_type": int(Grl.MediaType.AUDIO),
-            'location_filter': self._tracker_wrapper.location_filter(),
-            'media_ids': media_ids
-        }
+            }}
+        }} ORDER BY ?title ?artist
+        """.split())
 
         return query
 
@@ -520,50 +525,9 @@ class GrlTrackerWrapper(GObject.GObject):
                     self._songs_model.get_n_items(), 0, songs_added)
                 songs_added.clear()
 
-        query = """
-        SELECT
-            ?type ?urn ?title ?id ?url
-            ?artist ?album
-            ?duration ?trackNumber
-            ?albumDiscNumber
-            nie:usageCounter(?urn) AS ?playCount
-            ?tag AS ?favorite
-        WHERE {
-            SERVICE <dbus:%(miner_fs_busname)s> {
-                GRAPH tracker:Audio {
-                    SELECT
-                        %(media_type)s AS ?type
-                        ?song AS ?urn
-                        nie:title(?song) AS ?title
-                        ?song AS ?id
-                        nie:isStoredAs(?song) AS ?url
-                        nmm:artistName(nmm:artist(?song)) AS ?artist
-                        nie:title(nmm:musicAlbum(?song)) AS ?album
-                        nfo:duration(?song) AS ?duration
-                        nmm:trackNumber(?song) AS ?trackNumber
-                        nmm:setNumber(nmm:musicAlbumDisc(?song))
-                            AS ?albumDiscNumber
-                    WHERE {
-                        ?song a nmm:MusicPiece .
-                        %(location_filter)s
-                    }
-                    ORDER BY ?title
-                }
-            }
-            OPTIONAL {
-                ?urn nao:hasTag ?tag .
-                FILTER (?tag = nao:predefined-tag-favorite)
-            }
-        }
-        """.replace('\n', ' ').strip() % {
-            "miner_fs_busname": self._tracker_wrapper.props.miner_fs_busname,
-            "media_type": int(Grl.MediaType.AUDIO),
-            'location_filter': self._tracker_wrapper.location_filter()
-        }
-
         self.props.source.query(
-            query, self._METADATA_SONG_FILL_KEYS, self._fast_options,
-            _add_to_model)
+            self._song_media_query(), self._METADATA_SONG_FILL_KEYS,
+            self._fast_options, _add_to_model)
 
     def _initial_albums_fill(self) -> None:
         self._notificationmanager.push_loading()
@@ -703,11 +667,11 @@ class GrlTrackerWrapper(GObject.GObject):
             query, metadata_keys, self._fast_options, _add_to_artists_model)
 
     def get_artist_albums(
-            self, media: Grl.Source, model: Gfm.FilterListModel) -> None:
+            self, media: Grl.Source, model: Gtk.FilterListModel) -> None:
         """Get all albums by an artist
 
         :param Grl.Media media: The media with the artist id
-        :param Gfm.FilterListModel model: The model to fill
+        :param Gtk.FilterListModel model: The model to fill
         """
         self._notificationmanager.push_loading()
         artist_id = media.get_id()
@@ -745,7 +709,7 @@ class GrlTrackerWrapper(GObject.GObject):
             'location_filter': self._tracker_wrapper.location_filter()
         }
 
-        albums: List[Grl.Media] = []
+        albums: List[str] = []
 
         def query_cb(
                 source: Grl.Source, op_id: int, media: Optional[Grl.Media],
@@ -756,29 +720,26 @@ class GrlTrackerWrapper(GObject.GObject):
                 return
 
             if not media:
-                model.set_filter_func(albums_filter, albums)
+                custom_filter = Gtk.CustomFilter()
+                custom_filter.set_filter_func(albums_filter, albums)
+                model.set_filter(custom_filter)
                 self._notificationmanager.pop_loading()
                 return
 
-            albums.append(media)
+            albums.append(media.get_id())
 
-        def albums_filter(
-                corealbum: CoreAlbum, albums: List[Grl.Media]) -> bool:
-            for media in albums:
-                if media.get_id() == corealbum.props.media.get_id():
-                    return True
-
-            return False
+        def albums_filter(corealbum: CoreAlbum, albums: List[str]) -> bool:
+            return corealbum.props.media.get_id() in albums
 
         self.props.source.query(
             query, [Grl.METADATA_KEY_TITLE], self._fast_options, query_cb)
 
     def get_album_discs(
-            self, media: Grl.Media, disc_model: Gfm.SortListModel) -> None:
+            self, media: Grl.Media, disc_model: Gtk.SortListModel) -> None:
         """Get all discs of an album
 
         :param Grl.Media media: The media with the album id
-        :param Gfm.SortListModel disc_model: The model to fill
+        :param Gtk.SortListModel disc_model: The model to fill
         """
         self._notificationmanager.push_loading()
         album_id = media.get_id()
@@ -831,21 +792,16 @@ class GrlTrackerWrapper(GObject.GObject):
             query, [Grl.METADATA_KEY_ALBUM_DISC_NUMBER], self._fast_options,
             _disc_nr_cb)
 
-    def populate_album_disc_songs(
+    def get_album_disc(
             self, media: Grl.Media, disc_nr: int,
-            callback: Callable[
-                [Grl.Source, int, Optional[Grl.Media], int,
-                 Optional[GLib.Error]], None]) -> None:
-        # FIXME: Pass a model and fill it.
-        # FIXME: The query is similar to the other song queries, reuse
-        # if possible.
+            model: Gtk.FilterListModel) -> None:
         """Get all songs of an album disc
 
         :param Grl.Media media: The media with the album id
         :param int disc_nr: The disc number
-        :param callback: The callback to call for every song added
+        :param Gtk.FilterListModel model: The model to fill
         """
-        album_id: str = media.get_id()
+        album_id = media.get_id()
 
         query = """
         SELECT
@@ -909,8 +865,28 @@ class GrlTrackerWrapper(GObject.GObject):
             Grl.METADATA_KEY_URL
         ]
 
+        disc_song_ids: List[str] = []
+
+        def _filter_func(coresong: CoreSong) -> bool:
+            return coresong.props.grlid in disc_song_ids
+
+        def _callback(
+                source: Grl.Source, op_id: int, media: Grl.Media,
+                remaining: int, error: GLib.Error) -> None:
+            if error:
+                self._log.warning(f"Error: {error.domain}, {error.message}")
+                return
+
+            if media is None:
+                custom_filter = Gtk.CustomFilter()
+                custom_filter.set_filter_func(_filter_func)
+                model.set_filter(custom_filter)
+                return
+
+            disc_song_ids.append(media.get_source() + media.get_id())
+
         self.props.source.query(
-            query, metadata_keys, self._fast_options, callback)
+            query, metadata_keys, self._fast_options, _callback)
 
     def search(self, text: str) -> None:
         # FIXME: Searches are limited to not bog down the UI with
@@ -990,7 +966,9 @@ class GrlTrackerWrapper(GObject.GObject):
                 return
 
             if not media:
-                self._artists_search.set_filter_func(artist_filter)
+                custom_filter = Gtk.CustomFilter()
+                custom_filter.set_filter_func(artist_filter)
+                self._artists_search.set_filter(custom_filter)
                 self._notificationmanager.pop_loading()
                 return
 
@@ -1059,7 +1037,9 @@ class GrlTrackerWrapper(GObject.GObject):
                 return
 
             if not media:
-                self._albums_search.set_filter_func(album_filter)
+                custom_filter = Gtk.CustomFilter()
+                custom_filter.set_filter_func(album_filter)
+                self._albums_search.set_filter(custom_filter)
                 self._notificationmanager.pop_loading()
                 return
 
@@ -1133,7 +1113,9 @@ class GrlTrackerWrapper(GObject.GObject):
                 return
 
             if not media:
-                self._songs_search.set_filter_func(songs_filter)
+                custom_filter = Gtk.CustomFilter()
+                custom_filter.set_filter_func(songs_filter)
+                self._songs_search.set_filter(custom_filter)
                 self._notificationmanager.pop_loading()
                 return
 
